#!/bin/bash

#
#  hitlist_do_all.sh
#  Copyright (C) 2010 by USC/ISI
#  $Id$
#  
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License,
#  version 2, as published by the Free Software Foundation.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#  
#  
#  This code was originally written by Xun Fan <xunfan@isi.edu>
#  and Yuri Pradkin <yuri@isi.edu>.
#

# do all script for hitlist generation
set -o pipefail

# usage $0 <tracedir>
usage() {
    cat <<EOF

$0 [-yn] <tracedir>

Generate hitlists for census packaged in <tracedir>
Hitlists for censuses preceding that one must have been done.
Options:
   -y    answer 'y' to all prompts (overwrite everything)
   -n    answer 'n' to all prompts (overwrite nothing)
   -k    keep all intermediate files (do not delete)
EOF
}

ALL=
keepint=
while getopts "ynh" opt; do
    case $opt in
	( y ) ALL="y"; ;;
	( n ) ALL="n"; ;;
	( k ) keepint="y"; ;;
	( h|* ) usage; exit 1; ;;
    esac
done
shift $((OPTIND-1))

test $# -eq 1 || die "wrong number of parameters given. \nUsage: $0 <census trace directory>"

S=$(cd -P $(dirname $0); pwd)
tracedir=$1
CONFIG="$S/hitlist_config"
[ -e "$CONFIG" ] || die "configuration file $CONFIG does not exist"
. $CONFIG

HADOOP=$HADOOP_HOME/bin/hadoop
HDFS="$HADOOP_HOME/bin/hdfs dfs"

die () {
    echo "$@" 1>&2
    exit 1
}

overwrite_hdfs () {
    if $HDFS -test -e "$1" 2>/dev/null; then
	if [ -n "$ALL" ]; then
	    yn=$ALL
	    echo "HDFS $1 exists, overwrite (y/[n])? $yn"
	else
            read -p "HDFS $1 exists, overwrite (y/[n])? " yn
	fi
        if [ "$yn" != "y" ]; then
            echo "reusing existing $1"
            return 1
        else
            $HDFS -rmr $1 2>/dev/null
        fi
    fi
    return 0
}

overwrite () {
    if [ -e "$1" ]; then
	if [ -n "$ALL" ]; then
	    yn=$ALL
	    echo "File $1 exists, overwrite (y/[n])? $yn"
	else
            read -p "File $1 exists, overwrite (y/[n])? " yn
	fi
        if [ "$yn" != "y" ]; then
            echo "reusing existing $1"
            return 1
        else
            rm -rf $1
        fi
    fi
    return 0
}


#HITLIST_DEST_DIR may be a soft link
[ -d "$HITLIST_DEST_DIR" ] || mkdir -p "$HITLIST_DEST_DIR"
HITLIST_DEST_DIR=$( cd -P $HITLIST_DEST_DIR; pwd )

basetracedir=$(basename $tracedir)
#extract it<number><location> from name
tmp=${tracedir%-*}; base=${tmp##*_it}
dex=${base%[wce]}
letter=${base//[0-9]/}

CENSUS_NAME=$basetracedir
CENSUS_INDEX=$base #e.g. it40w  
CENSUS_INDEX_PREVIOUS=$((dex-1))$letter #e.g. it39w

[ -n "$HADOOP_HOME" ] || die "hadoop environment variable isn't set"
STREAMING_JAR=$HADOOP_HOME/build/hadoop-streaming.jar
INPUTFORMAT_JAR=$HADOOP_HOME/build/hadoop-icmptrain.jar
INPUTFORMAT_CLASS=edu.isi.hadoop.icmptrain.ICMPTrainInputFormat
export HADOOP_CLASSPATH=${INPUTFORMAT_JAR}

IANA_ALLOC=$S/tmp_hitlist_it$base
[ -e $tracedir/iana-*.fsdb ] || die "IANA fsdb record is missing from $tracedir"
if overwrite $IANA_ALLOC; then
    mkdir $IANA_ALLOC
    $S/generate_base_ipv4_hitlist.pl $tracedir/iana-*.fsdb $IANA_ALLOC || die "generate_base... failed"
fi

HITLIST_DIR=$HITLIST_INTERMEDIATE_DIR
test -d $HITLIST_DIR || die "HITLIST_DIR directory $HITLIST_DIR doesn't exist"

#
# start work
#
echo start time: $(date) >> $S/internet_address_hitlist_it$CENSUS_INDEX.log

HDFS_TRACES="/user/traces"
HDFS_ADDRESS_DATASETS="$HDFS_TRACES/address_scans_datasets"
HDFS_CENSUS="$HDFS_ADDRESS_DATASETS/$CENSUS_NAME"
$HDFS -mkdir $HDFS_TRACES >& /dev/null && \
    $HDFS -chmod a+wt $HDFS_TRACES
$HDFS -mkdir $HDFS_ADDRESS_DATASETS >& /dev/null && \
    $HDFS -chmod a+wt $HDFS_ADDRESS_DATASETS
$HDFS -mkdir $HDFS_CENSUS >& /dev/null
JOBNAME=hitlist-$CENSUS_INDEX

# 1. CENSUS TO INTERMEDIATE
#  input:  HDFS of census NN
#  output: HDFS single_intermediate_itNN
OUT_CTOI=single_intermediate_it$CENSUS_INDEX
if overwrite_hdfs $OUT_CTOI; then
    INP_CTOI="$HDFS_CENSUS/data"
    $HDFS -test -e "$INP_CTOI" >/dev/null 2>&1 || { 
        $HDFS -put file://$HITLIST_CENSUS_DIR/$CENSUS_NAME/data "$INP_CTOI" || \
            die "Cannot put $INP_CTOI"
    }
    MAP_CTOI="$S/census_to_intermediate_map"
    test -e $MAP_CTOI || die "file $MAP_CTOI doesn't exist"
    RED_CTOI="$S/census_to_intermediate_red"
    test -e $RED_CTOI || die "file $RED_CTOI doesn't exist"
    OPTS_CTOI='-jobconf stream.recordreader.icmptrain.keys=probeaddr -jobconf stream.recordreader.icmptrain.vals=type_code'
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass1"
    $HADOOP jar $STREAMING_JAR -file $INPUTFORMAT_JAR $OPTS_CTOI $JN_OPT \
        -inputformat $INPUTFORMAT_CLASS -input $INP_CTOI -output $OUT_CTOI \
        -mapper $MAP_CTOI -reducer $RED_CTOI || die "hadoop output $OUT_CTOI failed"
fi

# 2. INTERMEDIATE JOIN
#  input:  HDFS single_intermediate_itNN from step 1
#  input:  intermediate_it{NN-1} either from HDFS or from $HITLIST_DIR
#  output: HDFS intermediate_itNN, copy to $HITLIST_DIR
OUT_IJOIN=intermediate_it$CENSUS_INDEX
if overwrite_hdfs $OUT_IJOIN; then
    INP_IJOIN_1="intermediate_it$CENSUS_INDEX_PREVIOUS"
    inp_ijoin_1="$HITLIST_DIR/$INP_IJOIN_1"
    if $HDFS -test -e $INP_IJOIN_1; then
        echo "using hdfs version of $inp_ijoin_1"
    else
        echo "copying $inp_ijoin_1 to $INP_IJOIN_1"
        test -d $inp_ijoin_1 || die "dir $inp_ijoin_1 doesn't exist"
        $HDFS -put $inp_ijoin_1 $INP_IJOIN_1 || die "cannot hdfs put $inp_ijoin_1 $INP_IJOIN_1"
    fi
    INT_IJOIN_2=$OUT_CTOI
    MAP_IJOIN="$S/intermediate_join_map"
    test -e $MAP_IJOIN || die "file $MAP_IJOIN doesn't exist"
    RED_IJOIN="$S/intermediate_join_red"
    test -e $RED_IJOIN || die "file $RED_IJOIN doesn't exist"
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass2"

    $HADOOP jar $STREAMING_JAR $JNOPT -input $INP_IJOIN_1 -input $INT_IJOIN_2 -output $OUT_IJOIN \
        -mapper $MAP_IJOIN -reducer $RED_IJOIN || die "hadoop output $OUT_IJOIN fails!"
fi
DIR_IJOIN="$HITLIST_DIR/$OUT_IJOIN"
if overwrite $DIR_IJOIN; then
    $HDFS -get $OUT_IJOIN $HITLIST_DIR || die "hadoop_get $OUT_IJOIN/* fails!"
fi

# 3. GENERATE IPV4 HISTORY DATASET
#  input:  HDFS intermediate_itNN  (hex_ip, hex_history)
#  output: HDFS ipv4_history_itNN  (hex_ip, str_ip, hex_history, score)
OUT_IPV4_HISTORY=ipv4_history_it$CENSUS_INDEX
if overwrite_hdfs $OUT_IPV4_HISTORY; then
    INP_IPV4=intermediate_it$CENSUS_INDEX
    MAP_IPV4="$S/ipv4_history_map"
    RED_IPV4="$S/ipv4_history_red"
    test -e $MAP_IPV4 || die "file $MAP_IPV4 doesn't exist"
    test -e $RED_IPV4 || die "file $RED_IPV4 doesn't exist"
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass3"

    $HADOOP jar $STREAMING_JAR $JNOPT \
	-input $INP_IPV4 -output $OUT_IPV4_HISTORY \
        -mapper $MAP_IPV4 -reducer $RED_IPV4 || die "hadoop output $OUT_IPV4_HISTORY fails!"
fi
#Package internet_address_history
HISTORY_DATE=${CENSUS_NAME##*-}
HIST_DIST="$HITLIST_DEST_DIR/internet_address_history_it$CENSUS_INDEX-$HISTORY_DATE"
if overwrite $HIST_DIST; then
    mkdir -p $HIST_DIST || die "mkdir $HIST_DIST failed"
    HIST_FSDB=$HIST_DIST/internet_address_history_it$CENSUS_INDEX-$HISTORY_DATE.fsdb
    tmp=${TMPDIR:-/tmp}/$OUT_IPV4_HISTORY
    rm -rf $tmp
    $HDFS -get $OUT_IPV4_HISTORY ${TMPDIR:-/tmp}/ || die "hadoop_getmerge $OUT_IPV4_HISTORY failed!"
    {
        echo "#fsdb -F t hex_ip ip history score"
        sort -m --parallel=600 $tmp/part-* 
    } | bzip2 > $HIST_FSDB.bz2 || die "Can't package $HIST_FSDB.bz2"
    cp -p $tracedir/iana-*.fsdb $HIST_DIST/
    rm -rf $tmp
    #additional fsdb files
    #make probed_prefixes.fsdb
    prev_census="$HITLIST_CENSUS_DIR/internet_address_history_it$CENSUS_INDEX_PREVIOUS-*"
    prev_probed_prefixes="$prev_census/probed_prefixes.fsdb"
    curr_probed_prefixes="$HIST_DIST/probed_prefixes.fsdb"
    [ -e $prev_probed_prefixes ] || die "$prev_probed_prefixes does not exist"
    {
	< $prev_probed_prefixes \
	    dbcol iana_prefix census_name census_index
	
	< $tracedir/iana-*.fsdb \
	    dbrow ' _probing eq "true" ' |\
	    dbcolcreate -e $basetracedir census_name |\
	    dbcolcreate -e $dex census_index |\
	    dbcol iana_prefix census_name census_index
    } | dbfilestripcomments \
	| dbcoldefine iana_prefix census_name census_index \
	> $curr_probed_prefixes && echo "Wrote $curr_probed_prefixes" >&2

    #make prefix_ranges.fsdb
    curr_prefix_ranges="$HIST_DIST/prefix_ranges.fsdb"
    histprefixes=$(dbcol iana_prefix < $curr_probed_prefixes \
	| dbsort iana_prefix \
	| dbrowuniq iana_prefix \
	| dbfilestripcomments)
    { 
	echo "#fsdb -F t iana_prefix census_first census_last index_first index_last censuses censuses_max continuous"
	for p in $histprefixes; do 
	    t=$(dbrow "_iana_prefix eq '$p'" <$curr_probed_prefixes \
		| dbsort -n census_index \
		| dbfilestripcomments \
		| sed -n '1 p; $ {p;=}')
	    read p1 c1 i1 p2 c2 i2 lc <<<$t
	    dc=$((i2-i1+1)) 
	    cont=true
	    [ $lc != $dc ] && cont=false
	    echo -e "$p1\t$c1\t$c2\t$i1\t$i2\t$lc\t$dc\t$cont"; 
	done
    } >$curr_prefix_ranges && echo "Wrote $curr_prefix_ranges" >&2 

    #make a bit-map fsdb file
    curr_history_bits="$HIST_DIST/history_bits.fsdb"
    < $curr_probed_prefixes dbsort census_index \
	| dbrowuniq census_index \
	| dbcolcreate -e- history_bit census_start_date \
	| dbroweval "_history_bit=$dex-_census_index; _census_start_date=_census_name; _census_start_date =~ s/.*-//" \
	| dbcol history_bit census_start_date census_name census_index \
	| dbsort -n history_bit \
	| dbfilestripcomments \
	| dbcoldefine history_bit census_start_date census_name census_index \
	> $curr_history_bits && echo "Wrote $curr_history_bits" >&2 
    
    echo "Prepared $HIST_DIST" >&2
fi

#Pick winners from IPv4 history to build a raw hitlist
#input:  (hex_ip, str_ip, hex_history, score)
#output: (hex_prefix, best_hex_ip, best_score, best_str_ip)
OUT_RAW_HL=hitlists_it$CENSUS_INDEX
if overwrite_hdfs $OUT_RAW_HL; then
    INP_RAW_HL=$OUT_IPV4_HISTORY
    #MAP_RAW_HL="org.apache.hadoop.mapred.lib.IdentityMapper"
    MAP_RAW_HL="$S/hitlists_map"
    RED_RAW_HL="$S/hitlists_red"
    #test -e $MAP_RAW_HL || die "file $MAP_RAW_HL doesn't exist"
    test -e $RED_RAW_HL || die "file $RED_RAW_HL doesn't exist"
    $HADOOP jar $STREAMING_JAR -input $INP_RAW_HL -output $OUT_RAW_HL \
        -mapper $MAP_RAW_HL -reducer $RED_RAW_HL || die "hadoop output $OUT_RAW_HL fails!"
fi

DIR_RAW_HL="$HITLIST_DIR/$OUT_RAW_HL"
if overwrite $DIR_RAW_HL; then
    $HDFS -get $OUT_RAW_HL $HITLIST_DIR || die "hadoop_get $OUT_RAW_HL/* fails!"
fi

# 5. UPDATE OLD_STABLE_HITLIST
OUT_UPDATEHL=updated_old_hitlists_it$CENSUS_INDEX_PREVIOUS
if overwrite_hdfs $OUT_UPDATEHL; then
    INP_UPDATEHL_2=intermediate_it$CENSUS_INDEX
    # first put the old_stable_hitlist_intermediate... to hdfs
    INP_UPDATEHL_1=old_stable_hitlists_intermediate_it$CENSUS_INDEX_PREVIOUS
    if overwrite_hdfs $INP_UPDATEHL_1; then
	test -d $HITLIST_DIR/$INP_UPDATEHL_1 || die "dir $HITLIST_DIR/$INP_UPDATEHL_1 doesn't exist"
	$HDFS -put $HITLIST_DIR/$INP_UPDATEHL_1 $INP_UPDATEHL_1 || die "HDFS -put $HITLIST_DIR/$INP_UPDATEHL_1 $INP_UPDATEHL_1 fails!"
    fi
    MAP_UPDATEHL="$S/old_hitlists_update_with_intermediate_map"
    RED_UPDATEHL="$S/old_hitlists_update_with_intermediate_red"
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass5"
    
    test -e $MAP_UPDATEHL || die "file $MAP_UPDATEHL doesn't exist"
    test -e $RED_UPDATEHL || die "file $RED_UPDATEHL doesn't exist"
    $HADOOP jar $STREAMING_JAR $JNOPT -input $INP_UPDATEHL_1 -input $INP_UPDATEHL_2 -output $OUT_UPDATEHL \
	-mapper $MAP_UPDATEHL -reducer $RED_UPDATEHL || die "hadoop output $OUT_UPDATEHL fails!"
fi

# 6. STABLE HITLIST
OUT_STABLEHL=stable_hitlists_it${CENSUS_INDEX_PREVIOUS}_with_it$CENSUS_INDEX
if overwrite_hdfs $OUT_STABLEHL; then
    INP_STABLEHL_1=$OUT_RAW_HL
    INP_STABLEHL_2=$OUT_UPDATEHL
    
    MAP_STABLEHL="$S/stable_hitlists_map"
    RED_STABLEHL="$S/stable_hitlists_red"
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass6"
    
    test -e $MAP_STABLEHL || die "file $MAP_STABLEHL doesn't exist"
    test -e $RED_STABLEHL || die "file $RED_STABLEHL doesn't exist"
    $HADOOP jar $STREAMING_JAR $JNOPT -input $INP_STABLEHL_1 -input $INP_STABLEHL_2 \
	-output $OUT_STABLEHL -mapper $MAP_STABLEHL -reducer $RED_STABLEHL \
	|| die "hadoop output $OUT_STABLEHL fails!"
fi

# 7. MAKE STABLE HITLIST INTERMEDIATE
OUT_STABLEHLI=old_stable_hitlists_intermediate_it$CENSUS_INDEX
if overwrite_hdfs $OUT_STABLEHLI; then
    INP_STABLEHLI=$OUT_STABLEHL
    MAP_STABLEHLI="$S/old_hitlists_transfer_map"
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass7"
    test -e $MAP_STABLEHLI || die "file $MAP_STABLEHLI doesn't exist"
    $HADOOP jar $STREAMING_JAR $JNOPT -input $INP_STABLEHLI -output $OUT_STABLEHLI \
	-mapper $MAP_STABLEHLI -reducer org.apache.hadoop.mapred.lib.IdentityReducer \
	|| die "hadoop output $OUT_STABLEHLI fails!"
fi

DIR_STABLEHLI="$HITLIST_DIR/$OUT_STABLEHLI"
if overwrite $DIR_STABLEHLI; then
    $HDFS -get $OUT_STABLEHLI $HITLIST_DIR || die "hadoop_get $OUT_STABLEHLI fails!"
fi

# 8. JOIN WITH IANA ALLOCATION
HITLIST_DATE=${CENSUS_NAME##*-}
OUT_JOIN_IANA="internet_address_hitlist_it$CENSUS_INDEX-$HITLIST_DATE"
INP_JOIN_IANA_1=iana_allocation_new
if overwrite_hdfs $OUT_JOIN_IANA; then
    # first put temp iana_alloc base hitlist to hdfs
    if overwrite_hdfs $INP_JOIN_IANA_1; then
	test -d $IANA_ALLOC || die "dir $IANA_ALLOC doesn't exist"
	$HDFS -put $IANA_ALLOC $INP_JOIN_IANA_1 || die "HDFS -put $IANA_ALLOC fails!"
    fi
    INP_JOIN_IANA_2=$OUT_STABLEHL
    MAP_JOIN_IANA="$S/stable_hitlists_map"
    RED_JOIN_IANA="$S/stable_hitlists_with_IANA_red"
    JNOPT="-D mapreduce.job.name=$JOBNAME-pass8"
    
    test -e $MAP_JOIN_IANA || die "file $MAP_JOIN_IANA doesn't exist"
    test -e $RED_JOIN_IANA || die "file $RED_JOIN_IANA doesn't exist"
    $HADOOP jar $STREAMING_JAR $JNOPT -input $INP_JOIN_IANA_1 -input $INP_JOIN_IANA_2 \
	-output $OUT_JOIN_IANA \
	-mapper $MAP_JOIN_IANA -reducer $RED_JOIN_IANA || die "hadoop output $OUT_JOIN_IANA fails!"
fi
HL_DIST="$HITLIST_DEST_DIR/internet_address_hitlist_it$CENSUS_INDEX-$HITLIST_DATE"
if overwrite $HL_DIST; then
    mkdir -p $HL_DIST || die "mkdir $HL_DIST failed"
    tmp=$HL_DIST/tmp
    $HDFS -getmerge $OUT_JOIN_IANA $tmp || die "hadoop_getmerge $OUT_JOIN_IANA failed!"
    test -e $S/hitlists_final_cut_prefix || die "file $S/hitlists_final_cut_prefix doesn't exist"
    $S/hitlists_final_cut_prefix <$tmp | \
	dbsort hexip >$HL_DIST/$OUT_JOIN_IANA.fsdb \
	|| die "cut prefix and sort $DIR_JOIN_IANA_DATA/$OUT_JOIN_IANA failed"
    bzip2 $HL_DIST/$OUT_JOIN_IANA.fsdb || die "bzip2 $HL_DIST/$OUT_JOIN_IANA.fsdb failed"
    rm -rf $tmp $HL_DIST/.*.crc # delete intermediate file
fi

#copy iana fsdb file
cp $HITLIST_CENSUS_DIR/$CENSUS_NAME/iana-*.fsdb $HL_DIST || echo "cp iana-*.fsdb from census dir fails!"

echo "the output hitlist is under $HL_DIST"

echo "finish time: $(date)" >> $S/internet_address_hitlist_it$CENSUS_INDEX.log

# delete temp files from hdfs
if [ -z "$keepint" ]; then
    rm -rf $IANA_ALLOC
    $HDFS -rm -r $OUT_CTOI
    $HDFS -rm -r $OUT_IJOIN
    $HDFS -rm -r $OUT_RAW_HL
    $HDFS -rm -r $OUT_UPDATEHL
    $HDFS -rm -r $OUT_IPV4_HISTORY
    $HDFS -rm -r $OUT_STABLEHL
    $HDFS -rm -r $INP_JOIN_IANA_1
    $HDFS -rm -r $OUT_JOIN_IANA
else
    cat <<EOF
Intermediate files kept:
------------------------
file: $IANA_ALLOC
hdfs: $OUT_CTOI
hdfs: $OUT_IJOIN
hdfs: $OUT_RAW_HL
hdfs: $OUT_UPDATEHL
hdfs: $OUT_IPV4_HISTORY
hdfs: $OUT_STABLEHL
hdfs: $INP_JOIN_IANA_1
hdfs: $OUT_JOIN_IANA
------------------------
EOF
fi
